"""Highlights onto a particular graph."""

from typing import Callable, Hashable

from networkx.drawing import layout
from nxviz import layouts, lines, utils
from matplotlib.patches import Circle, Rectangle
import matplotlib.pyplot as plt
from functools import partial, update_wrapper
import numpy as np
import pandas as pd
from copy import deepcopy

import networkx as nx


def node(
    G,
    node,
    layout_func,
    group_by,
    sort_by=None,
    # visual properties
    color="red",  # color
    radius=1,  # size
    alpha=1.0,  # transparency
    clone=False,
    cloned_node_layout_kwargs={},
):
    """Highlight one particular node."""
    nt = utils.node_table(G, group_by=group_by, sort_by=sort_by)
    pos = layout_func(nt, group_by=group_by, sort_by=sort_by)
    ax = plt.gca()
    zorder = max([_.zorder for _ in ax.get_children()])
    c = Circle(xy=pos[node], fc=color, radius=radius, zorder=zorder)
    ax.add_patch(c)
    if clone:
        pos_cloned = layout_func(
            nt, group_by=group_by, sort_by=sort_by, **cloned_node_layout_kwargs
        )
        c = Circle(xy=pos_cloned[node], fc=color, radius=radius, zorder=zorder + 1)
        ax.add_patch(c)


circos_node = partial(node, layout_func=layouts.circos, group_by=None)
update_wrapper(circos_node, node)
circos_node.__name__ = "highlights.circos_node"
parallel_node = partial(node, layout_func=layouts.parallel, group_by=None)
update_wrapper(parallel_node, node)
parallel_node.__name__ = "highlights.parallel_node"

arc_node = partial(node, layout_func=layouts.arc, group_by=None)
update_wrapper(arc_node, node)
arc_node.__name__ = "highlights.arc_node"


hive_node = partial(
    node,
    layout_func=layouts.hive,
    clone=True,
    cloned_node_layout_kwargs={"rotation": np.pi / 6},
)
update_wrapper(hive_node, node)
hive_node.__name__ = "highlights.hive_node"
matrix_node = partial(
    node,
    layout_func=layouts.matrix,
    group_by=None,
    clone=True,
    cloned_node_layout_kwargs={"axis": "y"},
)
update_wrapper(matrix_node, node)
matrix_node.__name__ = "highlights.matrix_node"


def edge(
    G,
    source,
    target,
    layout_func,
    line_func,
    group_by,
    sort_by,
    color="red",  # color
    lw=1.0,  # size
    alpha=1.0,  # transparency
    clone=False,
    cloned_node_layout_kwargs={},
    line_func_aes_kw={"zorder": 30, "fc": "none"},
    line_func_kwargs={},
):
    """Highlight one particular edge."""
    nt = utils.node_table(G, group_by=group_by, sort_by=sort_by)
    et = utils.edge_table(G).query("source == @source").query("target == @target")
    pos = layout_func(nt, group_by=group_by, sort_by=sort_by)

    line_func_kwargs = deepcopy(line_func_kwargs)
    line_func_kwargs.update(
        et=et,
        pos=pos,
        edge_color=pd.Series([color], index=et.index),
        alpha=pd.Series([alpha], index=et.index),
        lw=pd.Series([lw], index=et.index),
        aes_kw=line_func_aes_kw,
    )

    if clone:
        pos_cloned = layout_func(
            nt, group_by=group_by, sort_by=sort_by, **cloned_node_layout_kwargs
        )
        line_func_kwargs["pos_cloned"] = pos_cloned

    patches = line_func(**line_func_kwargs)
    ax = plt.gca()
    for patch in patches:
        ax.add_patch(patch)


circos_edge = partial(
    edge,
    layout_func=layouts.circos,
    line_func=lines.circos,
    group_by=None,
    sort_by=None,
)
update_wrapper(circos_edge, node)
circos_edge.__name__ = "highlights.circos_edge"

arc_edge = partial(
    edge,
    layout_func=layouts.arc,
    line_func=lines.arc,
    group_by=None,
    sort_by=None,
    line_func_aes_kw={"zorder": 1},
)
update_wrapper(arc_edge, node)
arc_edge.__name__ = "highlights.arc_edge"

hive_edge = partial(
    edge,
    layout_func=layouts.hive,
    line_func=lines.hive,
    sort_by=None,
    clone=True,
    cloned_node_layout_kwargs={"rotation": np.pi / 6},
)
update_wrapper(hive_edge, node)
hive_edge.__name__ = "highlights.hive_edge"


matrix_edge = partial(
    edge,
    layout_func=layouts.matrix,
    line_func=lines.matrix,
    group_by=None,
    sort_by=None,
    clone=True,
    cloned_node_layout_kwargs={"axis": "y"},
    line_func_aes_kw={},
)
update_wrapper(matrix_edge, node)
matrix_edge.__name__ = "highlights.matrix_edge"

parallel_edge = partial(
    edge,
    layout_func=layouts.parallel,
    line_func=lines.line,
    sort_by=None,
    line_func_aes_kw={"zorder": 1},
)
update_wrapper(parallel_edge, node)
parallel_edge.__name__ = "highlights.parallel_edge"


def matrix_row(
    G: nx.Graph,
    node: Hashable,
    group_by: Hashable = None,
    sort_by: Hashable = None,
    axis="x",
    color="red",
):
    """Highlight one row (or column) in the matrix plot."""
    nt = utils.node_table(G)
    pos = layouts.matrix(nt, group_by=group_by, sort_by=sort_by, axis=axis)
    x, y = pos[node]

    width = 2
    height = 2 * len(G)
    xy = x - 1, y + 1
    if axis == "y":
        width, height = height, width
        xy = x + 1, y - 1
    rectangle = Rectangle(
        xy=xy, width=width, height=height, fc=color, ec="none", alpha=0.3
    )
    ax = plt.gca()
    ax.add_patch(rectangle)



"""Functions for drawing edges.

In drawing edges, we need to know some pieces of information beforehand.

Firstly,
"""

from copy import deepcopy
from functools import partial, update_wrapper
from typing import Callable, Dict, Hashable

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd

from nxviz import encodings, lines
from nxviz.utils import edge_table

default_edge_kwargs = dict(facecolor="none", zorder=1)


def line_width(et: pd.DataFrame, lw_by: Hashable):
    """Default edge line width function."""
    if lw_by is not None:
        return encodings.data_linewidth(et[lw_by])
    return pd.Series([1] * len(et), name="lw")


def transparency(et: pd.DataFrame, alpha_by: Hashable):
    """Default edge line transparency function."""
    if alpha_by is not None:
        return encodings.data_transparency(et[alpha_by])
    return pd.Series([0.1] * len(et), name="alpha")


def edge_colors(et: pd.DataFrame, color_by: Hashable):
    """Default edge line color function."""
    if color_by:
        return encodings.data_color(et[color_by])
    return pd.Series(["black"] * len(et), name="color_by")


def draw(
    G: nx.Graph,
    pos: Dict[Hashable, np.ndarray],
    lines_func: Callable,
    color_by: Hashable = None,
    lw_by: Hashable = None,
    alpha_by: Hashable = None,
    ax=None,
    encodings_kwargs: Dict = {},
    **linefunc_kwargs,
):
    """Draw edges to matplotlib axes.

    ## Parameters

    - `G`: A NetworkX graph.
    - `pos`: A dictionary mapping for x,y coordinates of a node.
    - `lines_func`: One of the line drawing functions from `nxviz.lines`
    - `color_by`: Categorical or quantitative edge attribute key to color edges by.
    - `lw_by`: Quantitative edge attribute key to determine line width.
    - `alpha_by`: Quantitative edge attribute key to determine transparency.
    - `ax`: Matplotlib axes object to plot onto.
    - `encodings_kwargs`: A dictionary of kwargs
        to determine the visual properties of the edge.
    - `linefunc_kwargs`: All other keyword arguments passed in
        will be passed onto the appropriate linefunc.

    Special keyword arguments for `encodings_kwargs` include:

    - `lw_scale`: A scaling factor for all edges' line widths.
        Equivalent to multiplying all line widths by this number.
    - `alpha_scale`: A scaling factor for all edges' line transparencies.
        Equivalent to multiplying all alphas by this number.
        The default transparency is 0.1,
        so an alpha_scale of any number greater than or equal to 10
        will result in 100% opaque lines.

    Everything else passed in here will be passed
    to the matplotlib Patch constructor;
    see `nxviz.lines` for more information.
    """
    et = edge_table(G)
    if ax is None:
        ax = plt.gca()
    edge_color = edge_colors(et, color_by)
    encodings_kwargs = deepcopy(encodings_kwargs)
    lw = line_width(et, lw_by) * encodings_kwargs.pop("lw_scale", 1.0)
    alpha = transparency(et, alpha_by) * encodings_kwargs.pop("alpha_scale", 1.0)

    aes_kw = {"facecolor": "none"}
    aes_kw.update(encodings_kwargs)
    patches = lines_func(
        et,
        pos,
        edge_color=edge_color,
        alpha=alpha,
        lw=lw,
        aes_kw=aes_kw,
        **linefunc_kwargs,
    )
    for patch in patches:
        ax.add_patch(patch)


circos = partial(draw, lines_func=lines.circos)
line = partial(draw, lines_func=lines.line)
arc = partial(draw, lines_func=lines.arc)
hive = partial(draw, lines_func=lines.hive)
matrix = partial(draw, lines_func=lines.matrix)

update_wrapper(circos, draw)
circos.__name__ = "edges.circos"

update_wrapper(line, draw)
line.__name__ = "edges.line"

update_wrapper(arc, draw)
arc.__name__ = "edges.arc"

update_wrapper(hive, draw)
hive.__name__ = "edges.hive"

update_wrapper(matrix, draw)
matrix.__name__ = "edges.matrix"



"""Polar/cartesian conversions functions."""

from numpy import arctan2 as atan2
from numpy import cos, pi, sin, sqrt


def to_cartesian(r, theta, proper=False):
    """
    Converts polar r, theta to cartesian x, y.
    """

    if proper:
        theta = to_proper_radians(theta)
    x = r * cos(theta)
    y = r * sin(theta)

    return x, y


import numpy as np


def to_polar(x, y):
    """
    Converts cartesian x, y to polar r, theta.
    """
    theta = atan2(y, x)
    r = sqrt(x ** 2 + y ** 2)

    if theta < 0:
        theta += 2 * np.pi

    return r, theta


def to_proper_radians(theta):
    """
    Converts theta (radians) to be within -pi and +pi.
    """
    if theta > pi or theta < -pi:
        theta = theta % pi
    return theta


def to_proper_degrees(theta):
    """
    Converts theta (degrees) to be within -180 and 180.
    """
    if theta > 180 or theta < -180:
        theta = theta % 180
    return theta


def to_degrees(theta):
    """
    Converts theta in radians to theta in degrees.
    """

    theta = to_proper_radians(theta)
    return theta / pi * 180


def to_radians(theta):
    """
    Converts theta in degrees to theta in radians.
    """
    theta = to_proper_degrees(theta)
    return theta * pi / 180



"""Functions to encode data inside visual properties of nodes and edges."""
from functools import partial
from typing import Callable, Tuple

import numpy as np
import pandas as pd

from matplotlib.cm import get_cmap
from matplotlib.colors import ListedColormap, Normalize, BoundaryNorm
from palettable.colorbrewer import qualitative, sequential

from nxviz.utils import infer_data_family


def data_cmap(data: pd.Series) -> Tuple:
    """Return a colormap for data attribute.

    Returns both the cmap and data family.
    """
    data_family = infer_data_family(data)
    if data_family == "categorical":
        base_cmap = qualitative
        num_categories = max(len(data.unique()), 3)
        if num_categories > 12:
            raise ValueError(
                f"It appears you have >12 categories for the key {data.name}. "
                "Because it's difficult to discern >12 categories, "
                "and because colorbrewer doesn't have a qualitative colormap "
                "with greater than 12 categories, "
                "nxviz does not support plotting with >12 categories."
            )
        cmap = ListedColormap(base_cmap.__dict__[f"Set3_{num_categories}"].mpl_colors)
    elif data_family == "ordinal":
        cmap = get_cmap("viridis")
    elif data_family == "continuous":
        cmap = get_cmap("viridis")
    elif data_family == "divergent":
        cmap = get_cmap("bwr")
    return cmap, data_family


def continuous_color_func(val, cmap, data: pd.Series):
    """Return RGBA of a value.

    ## Parameters

    - `val`: Value to convert to RGBA
    - `cmap`: A Matplotlib cmap
    - `data`: Pandas series.
    """
    norm = Normalize(vmin=data.min(), vmax=data.max())
    return cmap(norm(val))


def divergent_color_func(val, cmap, data: pd.Series):
    """Return RGBA for divergent color func.

    Divergent colormaps are best made symmetric.
    Hence, vmin and vmax are set appropriately here.
    """
    vmin = min(data.min(), -data.max())
    vmax = max(data.max(), -data.min())
    norm = Normalize(vmin=vmin, vmax=vmax)
    return cmap(norm(val))


def discrete_color_func(val, cmap, data: pd.Series):
    """Return RGB corresponding to a value.

    ## Parameters

    - `val`: Value to convert to RGBA
    - `cmap`: A Matplotlib cmap
    - `data`: Pandas series.
    """
    colors = sorted(data.unique())
    return cmap.colors[colors.index(val)]


def ordinal_color_func(val, cmap, data):
    """Return RGB corresponding to an ordinal value.

    ## Parameters

    - `val`: Value to convert to RGBA
    - `cmap`: A Matplotlib cmap
    - `data`: Pandas series.
    """
    bounds = np.arange(data.min(), data.max())
    norm = BoundaryNorm(bounds, cmap.N)
    return cmap(norm(val))


def color_func(data: pd.Series) -> Callable:
    """Return a color function that takes in a value and returns an RGB(A) tuple.

    This will do the mapping to the continuous and discrete color functions.
    """
    cmap, data_family = data_cmap(data)
    func = discrete_color_func
    if data_family in ["continuous", "ordinal"]:
        func = continuous_color_func
    return partial(func, cmap=cmap, data=data)


def data_color(data: pd.Series) -> pd.Series:
    """Return iterable of colors for a given data.

    `cfunc` gives users the ability to customize the color mapping of a node.
    The only thing that we expect is that it takes in a value
    and returns a matplotlib-compatible RGB(A) tuple or hexadecimal value.
    """

    cfunc = color_func(data)
    return data.apply(cfunc)


def data_transparency(data: pd.Series) -> pd.Series:
    """Transparency based on value."""
    norm = Normalize(vmin=data.min(), vmax=data.max())
    return data.apply(norm)


def data_size(data: pd.Series) -> pd.Series:
    """Square root node size."""
    return data.apply(np.sqrt)


def data_linewidth(data: pd.Series) -> pd.Series:
    """Line width scales linearly with property (by default)."""
    return data



"""Top-level nxviz API."""
from .api import (
    arc,
    hive,
    circos,
    parallel,
    matrix,
    geo,
    ArcPlot,
    HivePlot,
    MatrixPlot,
    CircosPlot,
)

import warnings


warnings.warn(
    """
nxviz has a new API! Version 0.7.0 onwards, the old class-based API is being
deprecated in favour of a new API focused on advancing a grammar of network
graphics. If your plotting code depends on the old API, please consider
pinning nxviz at version 0.7.0, as the new API will break your old code.

To check out the new API, please head over to the docs at
https://ericmjl.github.io/nxviz/ to learn more. We hope you enjoy using it!

(This deprecation message will go away in version 1.0.)
"""
)



"""Annotation submodule."""
from functools import partial, update_wrapper
from typing import Dict, Hashable

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize
from matplotlib.patches import Patch, Rectangle

from nxviz import encodings, layouts, utils
from nxviz.geometry import circos_radius, item_theta
from nxviz.polcart import to_cartesian


def text_alignment(x, y):
    """
    Align text labels based on the x- and y-axis coordinate values.

    This function is used for computing the appropriate alignment of the text
    label.

    For example, if the text is on the "right" side of the plot, we want it to
    be left-aligned. If the text is on the "top" side of the plot, we want it
    to be bottom-aligned.

    :param x, y: (`int` or `float`) x- and y-axis coordinate respectively.
    :returns: A 2-tuple of strings, the horizontal and vertical alignments
        respectively.
    """
    if x == 0:
        ha = "center"
    elif x > 0:
        ha = "left"
    else:
        ha = "right"
    if y == 0:
        va = "center"
    elif y > 0:
        va = "bottom"
    else:
        va = "top"

    return ha, va


def circos_group(
    G,
    group_by: Hashable,
    radius: float = None,
    radius_offset: float = 1,
    ax=None,
    midpoint=True,
):
    """Text annotation of node grouping variable on a circos plot."""
    nt = utils.node_table(G)
    groups = nt.groupby(group_by).apply(lambda df: len(df)).sort_index()
    proportions = groups / groups.sum()
    starting_points = proportions.cumsum() - proportions
    if midpoint:
        starting_points += proportions / 2
    angles = starting_points * 360
    radians = angles.apply(lambda x: x / 360 * 2 * np.pi)

    if ax is None:
        ax = plt.gca()

    if radius is None:
        radius = circos_radius(len(G)) + radius_offset

    for label, theta in radians.to_dict().items():
        x, y = to_cartesian(radius, theta)
        ha, va = text_alignment(x, y)
        ax.annotate(label, xy=(x, y), ha=ha, va=va)


def hive_group(G, group_by, ax=None, offset=np.pi / 12):
    """Text annotation of hive plot groups."""
    nt = utils.node_table(G)
    groups = sorted(nt[group_by].unique())

    if ax is None:
        ax = plt.gca()

    for grp in groups:
        theta = item_theta(groups, grp) + offset
        radius = 2 * (8 + len(nt[nt[group_by] == grp]) + 1)
        x, y = to_cartesian(radius, theta)
        ha, va = text_alignment(x, y)
        ax.annotate(grp, xy=(x, y), ha=ha, va=va)


def arc_group(
    G, group_by, ax=None, midpoint=True, y_offset=-1, rotation=45, ha="right", va="top"
):
    """Annotate arc group."""
    if ax is None:
        ax = plt.gca()
    nt = utils.node_table(G)
    groups = nt.groupby(group_by).apply(lambda df: len(df)).sort_index()
    proportions = groups / groups.sum()
    starting_points = proportions.cumsum() - proportions
    if midpoint:
        starting_points += proportions / 2
    starting_points *= len(G) * 2

    for label, starting_point in starting_points.to_dict().items():
        x = starting_point
        y = y_offset
        ax.annotate(label, xy=(x, y), ha=ha, va=va, rotation=rotation)


def parallel_group(
    G, group_by, ax=None, y_offset=-0.3, rotation=45, ha="right", va="top"
):
    """Annotate parallel plot groups."""
    if ax is None:
        ax = plt.gca()
    nt = utils.node_table(G)
    # groups = nt.groupby(group_by).apply(lambda df: len(df)).sort_index()
    groups = sorted(nt[group_by].unique())

    for i, label in enumerate(groups):
        x = i * 4
        y = y_offset
        ax.annotate(label, xy=(x, y), ha=ha, va=va, rotation=rotation)
    ax.relim()


def matrix_group(G, group_by, ax=None, offset=-3.0, xrotation=0, yrotation=90):
    """Annotate matrix plot groups."""
    if ax is None:
        ax = plt.gca()
    nt = utils.node_table(G)
    group_sizes = nt.groupby(group_by).apply(lambda df: len(df))
    proportions = group_sizes / group_sizes.sum()
    midpoint = proportions / 2
    starting_positions = proportions.cumsum() - proportions
    label_positions = (starting_positions + midpoint) * len(G) * 2
    label_positions += 1

    for label, position in label_positions.to_dict().items():
        # Plot the x-axis labels
        y = offset
        x = position
        ax.annotate(label, xy=(x, y), ha="center", va="center", rotation=xrotation)

        # Plot the y-axis labels
        x = offset
        y = position
        ax.annotate(label, xy=(x, y), ha="center", va="center", rotation=yrotation)


def matrix_block(G, group_by, color_by=None, ax=None, alpha=0.1):
    """Annotate group blocks on a matrix plot.

    Most useful for highlighting the within- vs between-group edges.
    """
    nt = utils.node_table(G)
    group_sizes = nt.groupby(group_by).apply(lambda df: len(df)) * 2
    starting_positions = group_sizes.cumsum() + 1 - group_sizes
    starting_positions

    colors = pd.Series(["black"] * len(group_sizes), index=group_sizes.index)
    if color_by:
        color_data = pd.Series(group_sizes.index, index=group_sizes.index)
        colors = encodings.data_color(color_data)
    # Generate patches first
    patches = []
    for label, position in starting_positions.to_dict().items():
        xy = (position, position)
        width = height = group_sizes[label]

        patch = Rectangle(
            xy, width, height, zorder=20, alpha=alpha, facecolor=colors[label]
        )
        patches.append(patch)

    if ax is None:
        ax = plt.gca()
    # Then add patches in.
    for patch in patches:
        ax.add_patch(patch)


def colormapping(data, legend_kwargs: Dict = {}, ax=None):
    """Annotate node color mapping.

    If the color attribute is continuous, a colorbar will be added to the matplotlib figure.
    Otherwise, a legend will be added.
    """
    cmap, data_family = encodings.data_cmap(data)
    if ax is None:
        ax = plt.gca()
    if data_family == "continuous":
        norm = Normalize(vmin=data.min(), vmax=data.max())
        scalarmap = ScalarMappable(
            cmap=cmap,
            norm=norm,
        )
        fig = plt.gcf()
        fig.colorbar(scalarmap)
    else:
        labels = data.drop_duplicates().sort_values()
        cfunc = encodings.color_func(data)
        colors = labels.apply(cfunc)
        patchlist = []
        for color, label in zip(colors, labels):
            data_key = Patch(color=color, label=label)
            patchlist.append(data_key)
        kwargs = dict(
            loc="best",
            ncol=int(len(labels) / 2),
            # bbox_to_anchor=(0.5, -0.05),
        )
        kwargs.update(legend_kwargs)
        legend = plt.legend(handles=patchlist, **kwargs)
        ax.add_artist(legend)


def node_colormapping(
    G,
    color_by,
    legend_kwargs: Dict = {"loc": "upper right", "bbox_to_anchor": (0.0, 1.0)},
    ax=None,
):
    """Annotate node color mapping."""
    nt = utils.node_table(G)
    data = nt[color_by]
    colormapping(data, legend_kwargs, ax)


def edge_colormapping(
    G,
    color_by,
    legend_kwargs={"loc": "lower right", "bbox_to_anchor": (0.0, 0.0)},
    ax=None,
):
    """Annotate edge color mapping."""
    if ax is None:
        ax = plt.gca()
    et = utils.edge_table(G)
    data = et[color_by]
    colormapping(data, legend_kwargs, ax)


def node_labels(G, layout_func, group_by, sort_by, ax=None):
    """Annotate node labels."""
    if ax is None:
        ax = plt.gca()

    nt = utils.node_table(G)
    pos = layout_func(nt, group_by, sort_by)
    for node in G.nodes():
        ax.annotate(text=node, xy=pos[node], ha="center", va="center")


parallel_labels = partial(node_labels, layout_func=layouts.parallel, sort_by=None)
update_wrapper(parallel_labels, node_labels)
parallel_labels.__name__ = "annotate.parallel_labels"

hive_labels = partial(node_labels, layout_func=layouts.hive, sort_by=None)
update_wrapper(hive_labels, node_labels)
hive_labels.__name__ = "annotate.hive_labels"

arc_labels = partial(node_labels, layout_func=layouts.arc, group_by=None, sort_by=None)
update_wrapper(arc_labels, node_labels)
arc_labels.__name__ = "annotate.arc_labels"


matrix_labels = partial(
    node_labels, layout_func=layouts.matrix, group_by=None, sort_by=None
)
update_wrapper(matrix_labels, node_labels)
matrix_labels.__name__ = "annotate.matrix_labels"


circos_labels = partial(
    node_labels, layout_func=layouts.circos, group_by=None, sort_by=None
)
update_wrapper(circos_labels, node_labels)
circos_labels.__name__ = "annotate.circos_labels"



"""High level nxviz plotting API."""


from functools import partial, update_wrapper
from typing import Callable, Dict, Hashable

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np

from nxviz import edges, nodes
from nxviz.plots import aspect_equal, despine

# This docstring applies to all plotting functions in this module.
# docstring =


def base(
    G: nx.Graph,
    node_layout_func: Callable,
    edge_line_func: Callable,
    group_by: Hashable,
    sort_by: Hashable,
    node_color_by: Hashable = None,
    node_alpha_by: Hashable = None,
    node_size_by: Hashable = None,
    node_enc_kwargs: Dict = {},
    edge_color_by: Hashable = None,
    edge_lw_by: Hashable = None,
    edge_alpha_by: Hashable = None,
    edge_enc_kwargs: Dict = {},
    node_layout_kwargs: Dict = {},
    edge_line_kwargs: Dict = {},
):
    """High-level graph plotting function.

    ## Parameters

    ### Basic

    - `G`: A NetworkX Graph.

    ### Nodes

    - `group_by`: Node metadata attribute key to group nodes.
    - `sort_by`: Node metadata attribute key to sort nodes.
    - `node_color_by`: Node metadata attribute key to color nodes.
    - `node_alpha_by`: Node metadata attribute key to set node transparency.
    - `node_size_by`: Node metadata attribute key to set node size.
    - `node_enc_kwargs`: Keyword arguments to set node visual encodings.
        TODO: Elaborate on what these arguments are.

    ### Edges

    - `edge_color_by`: Edge metdata attribute key to color edges.
    - `edge_lw_by`: Edge metdata attribute key to set edge line width.
    - `edge_alpha_by`: Edge metdata attribute key to set edge transparency.
    - `edge_enc_kwargs`: Keyword arguments to set node visual encodings.
        TODO: Elaborate on what these arguments are.
    """
    pos = node_layout_func(
        G,
        group_by=group_by,
        sort_by=sort_by,
        color_by=node_color_by,
        size_by=node_size_by,
        alpha_by=node_alpha_by,
        encodings_kwargs=node_enc_kwargs,
        layout_kwargs=node_layout_kwargs,
    )
    edge_line_func(
        G,
        pos,
        color_by=edge_color_by,
        lw_by=edge_lw_by,
        alpha_by=edge_alpha_by,
        encodings_kwargs=edge_enc_kwargs,
    )

    despine()
    aspect_equal()
    return plt.gca()


arc = partial(
    base,
    node_layout_func=nodes.arc,
    edge_line_func=edges.arc,
    group_by=None,
    sort_by=None,
)
update_wrapper(arc, base)
arc.__name__ = "api.arc"

circos = partial(
    base,
    node_layout_func=nodes.circos,
    edge_line_func=edges.circos,
    group_by=None,
    sort_by=None,
)
update_wrapper(circos, base)
circos.__name__ = "api.circos"

parallel = partial(
    base,
    node_layout_func=nodes.parallel,
    edge_line_func=edges.line,
    sort_by=None,
    node_enc_kwargs={"size_scale": 0.5},
)
update_wrapper(parallel, base)
parallel.__name__ = "api.parallel"

geo = partial(
    base,
    node_layout_func=nodes.geo,
    edge_line_func=edges.line,
    group_by=None,
    sort_by=None,
    node_enc_kwargs={"size_scale": 0.0015},
)
update_wrapper(geo, base)
geo.__name__ = "api.geo"


def base_cloned(
    G,
    node_layout_func,
    edge_line_func,
    group_by,
    sort_by=None,
    node_color_by=None,
    node_alpha_by=None,
    node_size_by=None,
    node_enc_kwargs={},
    edge_color_by=None,
    edge_lw_by=None,
    edge_alpha_by=None,
    edge_enc_kwargs={},
    node_layout_kwargs: Dict = {},
    edge_line_kwargs: Dict = {},
    cloned_node_layout_kwargs: Dict = {},
):
    """High-level graph plotting function.

    ## Parameters

    ### Basic

    - `G`: A NetworkX Graph.

    ### Nodes

    - `group_by`: Node metadata attribute key to group nodes.
    - `sort_by`: Node metadata attribute key to sort nodes.
    - `node_color_by`: Node metadata attribute key to color nodes.
    - `node_alpha_by`: Node metadata attribute key to set node transparency.
    - `node_size_by`: Node metadata attribute key to set node size.
    - `node_enc_kwargs`: Keyword arguments to set node visual encodings.
        TODO: Elaborate on what these arguments are.

    ### Edges

    - `edge_color_by`: Edge metdata attribute key to color edges.
    - `edge_lw_by`: Edge metdata attribute key to set edge line width.
    - `edge_alpha_by`: Edge metdata attribute key to set edge transparency.
    - `edge_enc_kwargs`: Keyword arguments to set node visual encodings.
        TODO: Elaborate on what these arguments are.
    """
    pos = node_layout_func(
        G,
        group_by=group_by,
        sort_by=sort_by,
        color_by=node_color_by,
        size_by=node_size_by,
        alpha_by=node_alpha_by,
        encodings_kwargs=node_enc_kwargs,
        layout_kwargs=node_layout_kwargs,
    )
    pos_cloned = node_layout_func(
        G,
        group_by=group_by,
        sort_by=sort_by,
        color_by=node_color_by,
        size_by=node_size_by,
        alpha_by=node_alpha_by,
        encodings_kwargs=node_enc_kwargs,
        layout_kwargs=cloned_node_layout_kwargs,
    )
    edge_line_func(
        G,
        pos,
        pos_cloned=pos_cloned,
        color_by=edge_color_by,
        lw_by=edge_lw_by,
        alpha_by=edge_alpha_by,
        encodings_kwargs=edge_enc_kwargs,
        **edge_line_kwargs,
    )

    despine()
    aspect_equal()
    return plt.gca()


hive = partial(
    base_cloned,
    node_layout_func=nodes.hive,
    edge_line_func=edges.hive,
    cloned_node_layout_kwargs={"rotation": np.pi / 6},
)
update_wrapper(hive, base_cloned)
hive.__name__ = "api.hive"

matrix = partial(
    base_cloned,
    group_by=None,
    node_layout_func=nodes.matrix,
    edge_line_func=edges.matrix,
    cloned_node_layout_kwargs={"axis": "y"},
)
update_wrapper(matrix, base_cloned)
matrix.__name__ = "api.matrix"


# Object-oriented API below, placed for compatibility.


class BasePlot:
    """Base Plot class."""

    def __init__(
        self,
        G: nx.Graph = None,
        node_grouping: Hashable = None,
        node_order: Hashable = None,
        node_color: Hashable = None,
        node_alpha: Hashable = None,
        node_size: Hashable = None,
        nodeprops: Dict = None,
        edge_color: Hashable = None,
        edge_alpha: Hashable = None,
        edge_width: Hashable = None,
        edgeprops: Dict = None,
    ):
        """Instantiate a plot.

        ## Parameters:

        - `G`: NetworkX graph to plot.
        - `node_grouping`: The node attribute on which to specify the grouping position of nodes.
        - `node_order`: The node attribute on which to specify the coloring of nodes.
        - `node_color`: The node attribute on which to specify the colour of nodes.
        - `node_alpha`: The node attribute on which to specify the transparency of nodes.
        - `node_size`: The node attribute on which to specify the size of nodes.
        - `nodeprops`: A `matplotlib`-compatible `props` dictionary.
        - `edge_color`: The edge attribute on which to specify the colour of edges.
        - `edge_alpha`: The edge attribute on which to specify the transparency of edges.
        - `edge_width`: The edge attribute on which to specify the width of edges.
        - `edgeprops`: A `matplotlib-compatible `props` dictionary.
        """
        import warnings

        warnings.warn(
            "As of nxviz 0.7, the object-oriented API is being deprecated "
            "in favour of a functional API. "
            "Please consider switching your plotting code! "
            "The object-oriented API wrappers remains in place "
            "to help you transition over. "
            "A few changes between the old and new API exist; "
            "please consult the nxviz documentation for more information. "
            "When the 1.0 release of nxviz happens, "
            "the object-oriented API will be dropped entirely."
        )

    def draw():
        """No longer implemented!"""
        pass


functional_api_names = [
    "group_by",
    "sort_by",
    "node_color_by",
    "node_alpha_by",
    "node_size_by",
    "node_enc_kwargs",
    "edge_color_by",
    "edge_alpha_by",
    "edge_lw_by",
    "edge_enc_kwargs",
]

object_api_names = [
    "node_grouping",
    "node_order",
    "node_color",
    "node_alpha",
    "node_size",
    "nodeprops",
    "edge_color",
    "edge_alpha",
    "edge_width",
    "edgeprops",
]

functional_to_object = dict(zip(functional_api_names, object_api_names))
object_to_functional = dict(zip(object_api_names, functional_api_names))


class ArcPlot(BasePlot):
    """Arc Plot."""

    def __init__(self, G, **kwargs):
        super().__init__()
        func_kwargs = {object_to_functional[k]: v for k, v in kwargs.items()}
        self.fig = plt.figure()
        self.ax = arc(G, **func_kwargs)


class CircosPlot(BasePlot):
    """Circos Plot."""

    def __init__(self, G, **kwargs):
        super().__init__()
        func_kwargs = {object_to_functional[k]: v for k, v in kwargs.items()}
        self.fig = plt.figure()
        self.ax = circos(G, **func_kwargs)


class HivePlot(BasePlot):
    """Hive Plot."""

    def __init__(self, G, **kwargs):
        super().__init__()
        func_kwargs = {object_to_functional[k]: v for k, v in kwargs.items()}
        self.fig = plt.figure()
        self.ax = hive(G, **func_kwargs)


class MatrixPlot(BasePlot):
    """Matrix Plot."""

    def __init__(self, G, **kwargs):
        super().__init__()
        func_kwargs = {object_to_functional[k]: v for k, v in kwargs.items()}
        self.fig = plt.figure()
        self.ax = matrix(G, **func_kwargs)
